from __future__ import annotations

import csv
import json
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    TextIO,
    Type,
)

from .filing import Filing

_formatters: Dict[str, Callable[[str], Formatter]] = {}
"""
A static container for Formatter factories.
"""


class FormatterException(Exception):
    pass


def get_formatter(name: str, destination: str) -> Optional[Formatter]:
    """
    Retrieve the registered formatter with the given name, or return
    `None` if it wasn't found.
    """
    constructor = _formatters.get(name, None)
    if constructor is None:
        return None
    return constructor(destination)


def register(name: str, factory: Callable[[str], Formatter]) -> None:
    """
    Register a new formatter with the given name. This allows it
    to be specified on the command line by its name.
    """
    _formatters[name] = factory


def registered_formatters() -> List[str]:
    """
    Return the names of all registered formatters.
    """
    return list(_formatters.keys())


class Formatter:
    """
    An abstract class that defines the structure of an object that
    can be used to output one or more filings for human or machine
    consumption.
    """

    def prologue(self) -> None:
        """
        A method that will be called before any filings are written
        to give the formatter a chance to prepare its output.
        """
        raise NotImplementedError()

    def write(self, filing: Filing) -> None:
        """
        A method called to output a single filing.
        """
        raise NotImplementedError()

    def write_all(self, filings: Iterable[Filing]) -> None:
        """
        A method called to output multiple filings.
        """
        for filing in filings:
            self.write(filing)

    def epilogue(self) -> None:
        """
        A method that will be called after all filings have been
        written to give the formatter a chance to close out its
        output.
        """
        raise NotImplementedError()


def register_formatter(
    name: str,
) -> Callable[[Type[Formatter]], Type[Formatter]]:
    """
    A decorator that allows a class to be registered as a formatter.
    Registered formatters must have a one-parameter constructor that
    accepts a string. This will be either the destination that was
    specified for the formatter or `None` if there was no destination
    specified. The formatter must work reasonably in either case.
    """

    def _decorator(cls: Type[Formatter]) -> Type[Formatter]:
        register(name, cls)
        return cls

    return _decorator


@register_formatter("file")
class FileFormatter(Formatter):
    """
    A formatter that writes to a file-like object.

    >>> f = FileFormatter(":stdout:")
    >>> f.prologue()
    >>> f.epilogue()
    """

    _file: Optional[TextIO] = None

    _has_written: bool = False

    _path: str

    def __init__(self, destination: str):
        self._path = destination

    def prologue(self) -> None:
        if self._file is not None:
            return

        if self._path == ":stdout:":
            self._file = sys.stdout
        elif self._path == ":stderr:":
            self._file = sys.stderr
        else:
            self._file = open(self._path, "w")

    def write(self, filing: Filing) -> None:
        if self._file is None:
            raise FormatterException("must call prologue() before write()")

        if self._has_written:
            self._file.write("\n")
        else:
            self._has_written = True

        # TODO: Come up with a way to let the user specify which fields to use
        self._file.write(f"Business Name:          {filing.business_name}\n")
        self._file.write(f"Formation Year:         {filing.formation_year}\n")
        self._file.write(
            f"Principal Officer Name: {filing.principal_officer_name}\n"
        )
        self._file.write(f"Website Address:        {filing.website_address}\n")

    def epilogue(self) -> None:
        if self._file is None:
            raise FormatterException("must call prologue() before epilogue()")

        if self._path != ":stdout:" and self._path != ":stderr:":
            self._file.close()

        self._file = None


@register_formatter("json")
class JSONFormatter(Formatter):
    """
    A formatter that outputs JSON-encoded filings.

    >>> f = JSONFormatter(":stdout:")
    >>> f.prologue()
    {"filings": [
    >>> f.epilogue()
    ]}

    """

    _file: Optional[TextIO] = None

    _has_written: bool = False

    _path: str

    def __init__(self, destination: str):
        self._path = destination

    def prologue(self) -> None:
        if self._file is not None:
            return

        if self._path == ":stdout:":
            self._file = sys.stdout
        elif self._path == ":stderr:":
            self._file = sys.stderr
        else:
            self._file = open(self._path, "w")

        self._file.write('{"filings": [')

    def write(self, filing: Filing) -> None:
        if self._file is None:
            raise FormatterException("must call prologue() before write()")

        if self._has_written:
            self._file.write(",")
        else:
            self._has_written = True

        json.dump(filing.to_json(), self._file)

    def epilogue(self) -> None:
        if self._file is None:
            raise FormatterException("must call prologue() before epilogue()")

        self._file.write("]}\n")

        if self._path != ":stdout:" and self._path != ":stderr:":
            self._file.close()

        self._file = None


class CSVWriterProtocol(Protocol):
    """
    A helper to let us type annotate the CSVFormatter.
    """

    def writerow(self, row: Iterable[Any]) -> None:
        ...


@register_formatter("csv")
class CSVFormatter(Formatter):
    """
    A Formatter implementation that creates CSV-formatted files.
    """

    _file: Optional[TextIO] = None

    _has_written: bool = False

    _path: str

    _writer: Optional[CSVWriterProtocol] = None

    def __init__(self, destination: str):
        self._path = destination

    def prologue(self) -> None:
        if self._file is not None:
            return

        if self._path == ":stdout:":
            self._file = sys.stdout
        elif self._path == ":stderr:":
            self._file = sys.stderr
        else:
            self._file = open(self._path, "w")

        self._writer = csv.writer(self._file)

        # TODO: Write column headers? How do we choose?

    def write(self, filing: Filing) -> None:
        if self._file is None:
            return

        if self._has_written:
            self._file.write(",")
        else:
            self._has_written = True

        # TODO: Write data from filing, choosing the same fields as prologue

    def epilogue(self) -> None:
        if self._file is None:
            return

        if self._path not in (":stdout:", ":stderr:"):
            self._file.close()

        self._file = None
        self._writer = None



from __future__ import annotations

from logging import getLogger
from typing import Dict, Optional, TextIO

from defusedxml import ElementTree

_logger = getLogger(__name__)


class Querier:
    """
    A Querier is a helper for extracting data from a filing XML file. It
    provides helper methods for getting typed data using XPaths.

    >>> querier = Querier.from_path("fixtures/filing.xml")
    >>> querier.find_str("/IRS990/FormationYr")
    '2004'
    >>> querier.find_int("/IRS990/TotalEmployeeCnt")
    19
    """

    _namespaces: Dict[str, str]

    _tree: ElementTree

    @staticmethod
    def from_path(xml_path: str) -> Querier:
        """
        A factory primarily intended for testing and interactive (REPL or
        Jupyter) use.
        """
        _logger.debug(f"create querier for '{xml_path}'")
        with open(xml_path) as xml_file:
            return Querier.from_file(xml_file)

    @staticmethod
    def from_file(xml_file: TextIO) -> Querier:
        """
        The standard factory which accepts a file-like object.
        """
        content = xml_file.read()

        # The XML files seem to come back with some kind of unicode identifier
        # sequence at the beginning and Python's XML processing barfs on it. So
        # here we strip it off to make sure that "<" (opening bracket for the
        # doctype) is the first character.
        while not content.startswith("<"):
            content = content[1:]

        byte_content = content.encode("utf-8")
        _logger.debug(f"hydrate querier: '{str(byte_content[:40])}'")

        tree = ElementTree.fromstring(byte_content)

        return Querier(tree)

    def __init__(self, tree: ElementTree):
        self._namespaces = {
            "efile": "http://www.irs.gov/efile",
        }
        self._tree = tree

    def find_float(self, path: str) -> Optional[float]:
        """
        Extract a field from the XML document adn convert
        it to a float before returning it.
        """
        _logger.debug(f"find float at '{path}'")

        raw_value = self.find_str(path)
        _logger.debug(f"found float - '{raw_value}'")

        if raw_value is None:
            return None

        return float(raw_value)

    def find_int(self, path: str) -> Optional[int]:
        """
        Extract a field from the XML document and
        convert it to an integer before returning it.
        """

        _logger.debug(f"find int at '{path}'")

        raw_value = self.find_str(path)
        _logger.debug(f"found int - '{raw_value}'")

        if raw_value is None:
            return None

        return int(raw_value)

    def find_str(self, path: str) -> Optional[str]:
        """
        Extract a field from the XML document and return
        it exactly as it was contained in the document.
        """

        _logger.debug(f"find str at '{path}'")

        full_path = self._add_namespace(path)
        _logger.debug(f"find str at '{full_path}'")

        element = self._tree.find(full_path, namespaces=self._namespaces)

        if element is None:
            _logger.debug(f"found str - '{None}'")
            return None

        element_content: str = element.text

        _logger.debug(f"found str - '{element_content}'")
        return element_content

    @staticmethod
    def _add_namespace(path: str) -> str:
        """
        Add the default namespace to all path segments
        and try to guess if we need to prepend any additional tags.
        """
        relative_path = path.lstrip("/")

        explicit_header = relative_path.startswith("ReturnHeader")
        explicit_data = relative_path.startswith("ReturnData")

        if not (explicit_header or explicit_data):
            # We assume ReturnData if nothing else was specified
            # because that's how the IRSx documentation formats
            # the paths.
            relative_path = f"ReturnData/{relative_path}"

        segments = [f"efile:{tag}" for tag in relative_path.split("/")]

        full_path = "/".join(segments)
        return full_path



from .annual_record import (
    ANNUAL_FIELD_NAMES,
    AnnualIndex,
    AnnualRecord,
    AnnualRecordDict,
    AnnualYear,
)
from .bmf_record import (
    BMF_FIELD_NAMES,
    BMFIndex,
    BMFRecord,
    BMFRecordDict,
    BMFRegion,
)
from .cache import Cache, DirectoryCache, MemoryCache
from .constants import (
    AWS_FILING_TEMPLATE,
    AWS_INDEX_TEMPLATE,
    DEFAULT_CACHE_PATH,
    IRS_BMF_TEMPLATE,
)
from .downloader import Downloader, DownloaderException, HTTPDownloader
from .filing import Filing
from .formatter import (
    FileFormatter,
    Formatter,
    FormatterException,
    JSONFormatter,
)
from .index import Index, IndexRecord
from .querier import Querier
from .result import Result



from __future__ import annotations

from logging import getLogger
from typing import (
    Callable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Union,
    cast,
)

from .annual_record import AnnualIndex, AnnualRecord, AnnualYear
from .bmf_record import BMFIndex, BMFRecord, BMFRegion
from .types import EIN

_logger = getLogger(__name__)


class IndexRecord(NamedTuple):
    """
    A composite index record that holds an AnnualRecord and a BMFRecord
    that both represent the same organization, as indicated by the EIN
    field. In all cases the two records here should have the same `ein`
    value.
    """

    annual_record: AnnualRecord
    bmf_record: BMFRecord

    @staticmethod
    def field_names() -> Set[str]:
        return set(AnnualRecord._fields + BMFRecord._fields)

    def get_field(self, name: str):
        if hasattr(self.annual_record, name):
            return getattr(self.annual_record, name)
        if hasattr(self.bmf_record, name):
            return getattr(self.bmf_record, name)
        raise KeyError(f"field '{name}' not found in record")

    def has_field(self, name: str):
        return hasattr(self.annual_record, name) or hasattr(
            self.bmf_record, name
        )


IndexFilter = Callable[[IndexRecord], bool]
"""
A callback that can specify whether a given record should be kept
as part of the collection of Filings produced by the index.
"""


class Index:
    """
    An Index provides a way to access a collection of filing documents that
    have been filtered in a reasonably efficient manner using the various
    fields in the annual and BMF (Business Master File) datasets.

    Years (annual) and regions (BMF) can be added to the index to set its
    initial pool of data. Filters, which further pare down the data actually
    fetched, may also be added at-will.

    Once the user wishes to fetch the collection of documents specified
    by the applied filters, the Index may be iterated directly to access
    each index record. The object ID necessary to fetch the document is
    provided on the annual index. Additionally, the object_ids() method may
    be used to access the object IDs directly.
    """

    _annual_indices: List[AnnualIndex] = []

    _annual_years: Set[AnnualYear] = set()

    _bmf_indices: List[BMFIndex] = []

    _bmf_regions: Set[BMFRegion] = set()

    _filters: Set[IndexFilter] = set()

    _length: Optional[int] = None

    def __iter__(self) -> Iterator[IndexRecord]:
        # TODO: Be smarter about which one to iterate first based on filters
        # TODO: Consider whether re-instantiating all these tuples is too slow
        for annual_index in self._annual_indices:
            for bmf_index in self._bmf_indices:
                for annual_record in annual_index.values():
                    ein = EIN(annual_record.ein)
                    if ein not in bmf_index:
                        _logger.debug(f"EIN not found in BMF index: {ein}")
                        continue

                    bmf_record = bmf_index[ein]

                    index_record = IndexRecord(annual_record, bmf_record)
                    passed = True
                    for cb in self._filters:
                        if not cb(index_record):
                            passed = False
                            break

                    if passed:
                        yield index_record

    def __len__(self) -> int:
        if self._length is not None:
            return self._length

        # TODO: This is stupid, don't instantiate the tuples, just count
        length = 0
        for _ in self:
            length += 1

        self._length = length
        return self._length

    def add_annual_index(
        self, year: AnnualYear, annual_index: AnnualIndex
    ) -> None:
        if year not in self._annual_years:
            self._length = None
            self._annual_indices.append(annual_index)
            self._annual_years.add(year)

    def add_bmf_index(self, region: BMFRegion, bmf_index: BMFIndex) -> None:
        if region not in self._bmf_regions:
            self._length = None
            self._bmf_indices.append(bmf_index)
            self._bmf_regions.add(region)

    def add_filter(self, cb: IndexFilter) -> None:
        if cb not in self._filters:
            self._length = None
            self._filters.add(cb)

    def object_ids(self) -> Iterator[str]:
        """
        Iterate over the object IDs that correspond to the filing documents
        reflected by this Index.
        """
        for record in self:
            yield record.annual_record.object_id

    def remove_filter(self, cb: IndexFilter) -> None:
        if cb in self._filters:
            self._length = None
            self._filters.remove(cb)

    def remove_index(self, index: Union[AnnualIndex, BMFIndex]) -> None:
        if index in self._annual_indices:
            self._length = None
            self._annual_indices.remove(cast(AnnualIndex, index))
        if index in self._bmf_indices:
            self._length = None
            self._bmf_indices.remove(cast(BMFIndex, index))



import os
from logging import getLogger
from typing import Container, Dict, Optional

from .constants import DEFAULT_CACHE_PATH

_logger = getLogger(__name__)


class Cache(Container):
    """
    A generic, abstract, cache to allow fast lookups for XML filings or
    CSV index files. This exists to keep us from having to download
    large or numerous files every time we update the site or data.

    TODO: Add a mechanism for invalidating / replacing cached documents
    """

    def get(self, cache_key: str) -> Optional[str]:
        """
        Fetch a document from the cache.
        """
        raise NotImplementedError()

    def put(self, cache_key: str, content: str) -> str:
        """
        Cache a document.
        """
        raise NotImplementedError()


class MemoryCache(Cache):
    """
    A naive, in-memory cache that does no eviction or invalidation.

    >>> c = MemoryCache()
    >>> c.get("foo")
    >>> c.put("foo", "bar")
    'bar'
    >>> c.get("foo")
    'bar'
    """

    _dict: Dict[str, str]

    def __contains__(self, __x: object) -> bool:
        return __x in self._dict

    def __init__(self):
        self._dict = {}

    def get(self, cache_key: str) -> Optional[str]:
        content = self._dict.get(cache_key, None)
        if content is None:
            _logger.debug(f"cache miss fetching cache key '{cache_key}'")
        _logger.debug(f"cache hit fetching cache key '{cache_key}'")
        return content

    def put(self, cache_key: str, content: str) -> str:
        _logger.debug(f"caching cache key '{cache_key}'")
        self._dict[cache_key] = content
        return content


class DirectoryCache(Cache):
    """
    A simple on-disk cache that saves documents to the given
    directory.

    TODO: Opened files need to be closed
    TODO: OMG so much edge case and error checking, seriously

    >>> import os.path as p
    >>> import tempfile as tf
    >>> d = tf.mkdtemp()
    >>> c = DirectoryCache(".xml", d)
    >>> c.get("foo")
    >>> c.put("foo", "bar")
    'bar'
    >>> p.exists(f"{d}/foo.xml")
    True
    """

    _extension: str

    _path: str

    def __contains__(self, __x: object) -> bool:
        pass

    def __init__(self, extension: str, path: str = DEFAULT_CACHE_PATH):
        if extension.startswith("."):
            self._extension = extension
        else:
            self._extension = f".{extension}"

        if path == "":
            # Use the default directory
            # TODO: Better error checking here, what if it's a file?
            if not os.path.isdir(DEFAULT_CACHE_PATH):
                os.mkdir(DEFAULT_CACHE_PATH)
            self._path = DEFAULT_CACHE_PATH
            return

        if os.path.exists(path):
            # Attempt to use the existing directory
            if not os.path.isdir(path):
                raise ValueError(f"path '{path}' is not a directory")

            self._path = path
            return

        os.mkdir(path)
        self._path = path

    def get(self, cache_key: str) -> Optional[str]:
        path = os.path.join(self._path, f"{cache_key}{self._extension}")
        if os.path.exists(path) and os.path.isfile(path):
            _logger.debug(f"cache hit fetching cache key '{cache_key}'")
            with open(path) as cache_file:
                return cache_file.read()
        _logger.debug(f"cache miss fetching cache key '{cache_key}'")
        return None

    def put(self, cache_key: str, content: str) -> str:
        _logger.debug(f"caching cache key '{cache_key}'")
        path = os.path.join(self._path, f"{cache_key}{self._extension}")
        with open(path, "w") as cache_file:
            cache_file.write(content)
        return content



from __future__ import annotations

from typing import Any, Dict, Iterable, Iterator

from .downloader import Downloader
from .filing import Filing, FilingFilter
from .index import Index
from .querier import Querier


class Result:
    """
    A lazily-evaluated query result based on the given index and downloader.

    The result can be filtered by providing a callback to the `Result.filter`
    method. This will return a new, independent `Result` that will apply all
    filters already applied by the previous `Result`, plus the one that was
    provided to the method.

    TODO: Enforce immutability assumption
    Right now this type is designed to be immutable so that a given Result
    always produces the same results and you can't accidentally change it
    during iteration, for example. But indices are no longer immutable
    because it was easier that way. However, we may need to go back to
    immutable indices, or copy the data we need from the index when it is
    used to create a Result so that the Result doesn't actually depend on
    the actual Index instance.
    """

    _downloader: Downloader

    _filters: Iterable[FilingFilter]

    _index: Index

    def __init__(
        self,
        downloader: Downloader,
        index: Index,
        filters: Iterable[FilingFilter] = (),
    ):
        self._downloader = downloader
        self._filters = filters
        self._index = index

    def __iter__(self) -> Iterator[Filing]:
        for record in self._index:
            xml_file = self._downloader.fetch(record.annual_record.object_id)
            querier = Querier.from_file(xml_file)
            filing = Filing(querier)
            passed = True
            for cb in self._filters:
                if not cb(filing):
                    passed = False
                    break
            if passed:
                yield filing

    def add_filter(self, cb: FilingFilter,) -> Result:
        """
        Apply a filter to the returned filings. Filings that don't pass the
        given function will not be yielded by the result.
        """
        return Result(
            self._downloader, self._index, list(self._filters) + [cb],
        )

    def download_count(self) -> int:
        """
        The number of filing documents that will be downloaded when this
        Result object is iterated. This may be different from the number
        of Filings actually produced upon iteration since some of them
        may be filtered out here (which requires downloading them first).
        """
        return len(self._index)

    def skip(self, n: int) -> Result:
        """
        Return a result that is identical to this one but doesn't contain the
        first `n` results.
        """
        # TODO: Implement me

    def take(self, n: int) -> Result:
        """
        Return a result that is identical to this one but contains only the
        first `n` results.
        """
        # TODO: Implement me

    def to_json(self) -> Dict[str, Any]:
        """
        Convert the result into JSON suitable for use as data for modules.

        TODO: Think about additional metadata we might want
        """
        return {
            "filings": [f.to_json() for f in self],
        }



from __future__ import annotations

import dataclasses as dc
import json
import logging
from argparse import ArgumentParser, Namespace
from datetime import datetime
from typing import Any, Dict, Iterable, List, Mapping, NamedTuple

from .cache import Cache, DirectoryCache, MemoryCache
from .constants import (
    CURRENT_VERSION,
    DEFAULT_CACHE_PATH,
    PROGRAM_DESCRIPTION,
    PROGRAM_NAME,
)
from .filing import Filing
from .formatter import (
    FileFormatter,
    Formatter,
    get_formatter,
    registered_formatters,
)
from .index import IndexRecord

_log_levels = {
    "none": -100,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARN,
    "error": logging.ERROR,
    "fatal": logging.FATAL,
}

_logger = logging.getLogger(__name__)


parser = ArgumentParser(
    prog=PROGRAM_NAME,
    description=PROGRAM_DESCRIPTION,
    epilog="A project of *Code for Montana*.",
)

parser.add_argument(
    "--version", "-v", action="version", version=CURRENT_VERSION,
)

parser.add_argument(
    "--use-disk-cache",
    action="store_true",
    default=False,
    help="cache index and filing documents to disk",
)

parser.add_argument(
    "--cache-path",
    type=str,
    default=DEFAULT_CACHE_PATH,
    help="path to use for reading and writing cache data",
)

parser.add_argument(
    "--dry-run",
    action="store_true",
    default=False,
    help="report how many documents would be "
    + "downloaded / process and nothing else",
)

parser.add_argument(
    "--load-filters",
    action="append",
    type=str,
    help="read filters from a JSON file",
)

parser.add_argument(
    "--log-level",
    type=str,
    choices=_log_levels.keys(),
    default="none",
    help="set the log level, no logging is done by default",
)

parser.add_argument(
    "--no-confirm",
    action="store_true",
    default=False,
    help="do not interactively confirm large downloads, for shell scripts",
)

parser.add_argument(
    "--save-filters", type=str, help="save the filters applied to a JSON file",
)

if "json" in registered_formatters():
    parser.add_argument(
        "--to-json",
        action="store_true",
        default=False,
        help="output extracted data to JSON, equivalent to --formatter=json",
    )

parser.add_argument(
    "--formatter",
    type=str,
    choices=registered_formatters(),
    default=registered_formatters()[0],
    help="output formatter, use --destination to specify file name",
)

parser.add_argument(
    "--destination",
    type=str,
    default=":stdout:",
    help="file to which output should be written (or ':stdout:', ':stderr:')",
)

parser.add_argument(
    "--regions",
    type=str,
    default="mt",
    help="regions to search, comma-separated "
    + "(two character state abbreviations, for most)",
)

parser.add_argument(
    "--years",
    type=str,
    default=":current:",
    help="years to search, comma-separated "
    + "(':current:' is the most recent completed year)",
)

# -------------------- #
# Index record filters #
# -------------------- #

for index_filter_field_name in IndexRecord.field_names():
    parser.add_argument(
        f"--{index_filter_field_name}",
        type=str,
        help=f"apply a filter to the {index_filter_field_name} index field",
    )

# -------------- #
# Filing filters #
# -------------- #

for filing_filter_field_name in dc.fields(Filing):
    name = filing_filter_field_name.name
    parser.add_argument(
        f"--{filing_filter_field_name.name}",
        type=str,
        help=f"apply a filter to the {name} filing field",
    )


class Options(NamedTuple):
    @staticmethod
    def from_args(args: Namespace) -> Options:
        if hasattr(args, "to_json") and args.to_json:
            formatter = get_formatter("json", args.destination)
        else:
            formatter = get_formatter(args.formatter, args.destination)
        if formatter is None:
            formatter = FileFormatter(args.destination)

        index_cache: Cache
        filing_cache: Cache
        if args.use_disk_cache:
            index_cache = DirectoryCache(".csv", args.cache_path)
            filing_cache = DirectoryCache(".xml", args.cache_path)
        else:
            index_cache = MemoryCache()
            filing_cache = MemoryCache()

        # Gather all the index filters
        index_filters: Dict[str, str] = {}
        for index_field_name in IndexRecord.field_names():
            if hasattr(args, index_field_name):
                value = getattr(args, index_field_name)
                if value is not None:
                    index_filters[index_field_name] = value
        _logger.debug(f"extracted index filters: {index_filters}")

        # Gather all the filing filters
        filing_filters: Dict[str, str] = {}
        for filing_field_name in dc.fields(Filing):
            name = filing_field_name.name
            if hasattr(args, name):
                value = getattr(args, name)
                if value is not None:
                    filing_filters[name] = value
        _logger.debug(f"extracted filing filters: {filing_filters}")

        # Check for JSON files specifying filters
        if args.load_filters is not None:
            for filters_path in args.load_filters:
                with open(filters_path) as filters_file:
                    filters_data: Dict[str, Any] = json.load(filters_file)

                # JSON file has two top-level keys, "index" and "filing"

                index_data: Dict[str, str] = filters_data["index"]
                if index_data is not None:
                    index_filters.update(index_data)

                filing_data: Dict[str, str] = filters_data["filing"]
                if filing_data is not None:
                    filing_filters.update(filing_data)

        # Save filters if we need to do so
        if args.save_filters is not None:
            payload = {
                "index": index_filters,
                "filing": filing_filters,
            }
            with open(args.save_filters, "w") as saveFile:
                json.dump(payload, saveFile)

        # Figure out the years we want
        years: List[str] = []
        years_str = args.years.split(",")
        for year_str in years_str:
            if year_str == ":current:":
                # We use last year since it is most likely to have
                # complete, or at least reasonably complete, data.
                # The other problem is that the current year probably
                # doesn't even exist until a couple months into the
                # year so the user would get mysterious 404s.
                years.append(str(datetime.today().year - 1))
                continue
            years.append(year_str)

        # Choose our regions
        regions: List[str] = []
        regions_str = args.regions.split(",")
        for region_str in regions_str:
            regions.append(region_str)

        return Options(
            dry_run=args.dry_run,
            formatter=formatter,
            filing_cache=filing_cache,
            index_cache=index_cache,
            log_level=_log_levels[args.log_level],
            log_level_human=args.log_level,
            no_confirm=args.no_confirm,
            filing_filters=filing_filters,
            index_filters=index_filters,
            to_json=args.to_json,
            regions=regions,
            years=years,
        )

    dry_run: bool

    formatter: Formatter

    filing_cache: Cache

    index_cache: Cache

    log_level: int

    log_level_human: str

    no_confirm: bool

    regions: Iterable[str]

    years: Iterable[str]

    filing_filters: Mapping[str, str] = {}

    index_filters: Mapping[str, str] = {}

    to_json: bool = False



from __future__ import annotations

import csv
from logging import getLogger
from typing import Any, Dict, Iterable, List, NamedTuple, NewType, TextIO

from .downloader import Downloader
from .types import EIN

_logger = getLogger(__name__)


BMF_FIELD_NAMES: List[str] = [
    "EIN",
    "NAME",
    "ICO",
    "STREET",
    "CITY",
    "STATE",
    "ZIP",
    "GROUP",
    "SUBSECTION",
    "AFFILIATION",
    "CLASSIFICATION",
    "RULING",
    "DEDUCTIBILITY",
    "FOUNDATION",
    "ACTIVITY",
    "ORGANIZATION",
    "STATUS",
    "TAX_PERIOD",
    "ASSET_CD",
    "INCOME_CD",
    "FILING_REQ_CD",
    "PF_FILING_REQ_CD",
    "ACCT_PD",
    "ASSET_AMT",
    "INCOME_AMT",
    "REVENUE_AMT",
    "NTEE_CD",
    "SORT_NAME",
]

BMFRegion = NewType("BMFRegion", str)

BMFRecordDict = NewType("BMFRecordDict", Dict[str, str])
"""
A dictionary presumed to have the keys listed in `BMF_FIELD_NAMES`
so that it can be used to create an instance of `BMFRecord`.
"""


class BMFRecord(NamedTuple):
    """
    A single record from a BMF (Business Master File) index.

    See https://www.irs.gov/charities-non-profits/exempt-organizations-business-master-file-extract-eo-bmf
    for more information about the data, including field descriptions.

    >>> from pyrs990.downloader import FakeDownloader
    >>> documents = {"mt": open("fixtures/bmf_index.csv")}
    >>> downloader = FakeDownloader(documents)
    >>> index0 = get_bmf_index(BMFRegion("mt"), downloader)
    >>> index1 = get_bmf_index(BMFRegion("mt"), downloader)
    >>> id(index0) == id(index1)
    True
    >>> len(index0)
    2
    >>> "010571299" in index0
    True
    """

    @staticmethod
    def from_csv(index_file: TextIO,) -> Iterable[BMFRecord]:
        records: List[BMFRecord] = []
        for row in csv.DictReader(index_file):
            record = BMFRecord.from_dict(row)
            records.append(record)
        return records

    @staticmethod
    def from_dict(d: Dict[str, str],) -> BMFRecord:
        params: Dict[str, str] = {}
        try:
            for name in BMF_FIELD_NAMES:
                params[name.lower()] = d[name]
        except KeyError as e:
            _logger.error(f"malformed index row - '{e}' - '{d}'")
            raise
        return BMFRecord(**params)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"BMFRecord(name='{self.name}', ein='{self.ein}')"

    def to_dict(self) -> Dict[str, Any]:
        """
        Support for JSON serialization.
        """
        return dict(self.__dict__)

    ein: str
    name: str
    ico: str
    street: str
    city: str
    state: str
    zip: str
    group: str
    subsection: str
    affiliation: str
    classification: str
    ruling: str
    deductibility: str
    foundation: str
    activity: str
    organization: str
    status: str
    tax_period: str
    asset_cd: str
    income_cd: str
    filing_req_cd: str
    pf_filing_req_cd: str
    acct_pd: str
    asset_amt: str
    income_amt: str
    revenue_amt: str
    ntee_cd: str
    sort_name: str


BMFIndex = Dict[EIN, BMFRecord]

_bmf_indices: Dict[BMFRegion, BMFIndex] = {}


def get_bmf_index(region: BMFRegion, downloader: Downloader) -> BMFIndex:
    if region in _bmf_indices:
        return _bmf_indices[region]

    index: BMFIndex = {}
    _bmf_indices[region] = index

    index_file = downloader.fetch(str(region))
    for record in BMFRecord.from_csv(index_file):
        index[EIN(record.ein)] = record
    return index



import sys

from . import app
from ._options import Options, _log_levels, parser

if __name__ == "__main__":
    import logging

    logging.basicConfig()
    logger = logging.getLogger(__package__)

    args = parser.parse_args(sys.argv[1:])

    # We do this kinda janky like without the Options instance
    # so that we can log inside the Options factory.
    if args.log_level != "none":
        logger.setLevel(_log_levels[args.log_level])

    options = Options.from_args(args)

    app.run(options)



from __future__ import annotations

from dataclasses import asdict, dataclass, field, fields
from logging import getLogger
from typing import Any, Callable, Dict, Optional

from .querier import Querier

_logger = getLogger(__name__)


def float_factory(path: str) -> Dict[str, Callable[[Querier], Optional[float]]]:
    """
    A helper that returns the correct metadata for a field that receives a
    float value and requires only a simple XPath query.
    """

    def _get_float(querier: Querier) -> Optional[float]:
        return querier.find_float(path)

    return {"factory": _get_float}


def int_factory(path: str) -> Dict[str, Callable[[Querier], Optional[int]]]:
    """
    A helper that returns the correct metadata for a field that receives a
    int value and requires only a simple XPath query.
    """

    def _get_int(querier: Querier) -> Optional[int]:
        return querier.find_int(path)

    return {"factory": _get_int}


def str_factory(path: str) -> Dict[str, Callable[[Querier], Optional[str]]]:
    """
    A helper that returns the correct metadata for a field that receives a
    str value and requires only a simple XPath query.
    """

    def _get_str(querier: Querier) -> Optional[str]:
        return querier.find_str(path)

    return {"factory": _get_str}


def two_part_str_factory(
    path1: str, path2: str, sep="\n"
) -> Dict[str, Callable[[Querier], Optional[str]]]:
    """
    A helper that returns a two-part string field with the parts
    split by the given separator.
    """

    def _factory(querier: Querier) -> Optional[str]:
        line1 = querier.find_str(path1)
        line2 = querier.find_str(path2)
        full: str = sep.join([l for l in [line1, line2] if l is not None])
        if full == "":
            return None
        return full

    return {"factory": _factory}


@dataclass(init=False)
class Filing:
    """
    Encapsulate a Filing as a semantic object with appropriately named
    fields.

    TODO: Write some doctests, build a Querier just like in its tests
    TODO: Explain how to add a new field
    TODO: Consider making this "frozen", see Python docs for implications

    >>> from pyrs990.querier import Querier
    >>> q = Querier.from_path("fixtures/filing.xml")
    >>> f = Filing(q)
    >>> f.formation_year
    2004
    >>> f.employee_count
    19
    >>> f.business_name
    'VOICE OF SAN DIEGO'
    >>> type(f.to_json())
    <class 'dict'>
    """

    activity_or_mission_description: Optional[str] = field(
        metadata=str_factory("/IRS990/ActivityOrMissionDesc")
    )
    business_name: Optional[str] = field(
        metadata=two_part_str_factory(
            "/ReturnHeader/Filer/BusinessName/BusinessNameLine1Txt",
            "/ReturnHeader/Filer/BusinessName/BusinessNameLine2Txt",
            sep=" ",
        )
    )
    formation_year: Optional[int] = field(
        metadata=int_factory("/IRS990/FormationYr")
    )
    gross_receipts: Optional[int] = field(
        metadata=int_factory("/IRS990/GrossReceiptsAmt")
    )
    employee_count: Optional[int] = field(
        metadata=int_factory("/IRS990/TotalEmployeeCnt")
    )
    principal_officer_name: Optional[str] = field(
        metadata=str_factory("/IRS990/PrincipalOfficerNm")
    )
    tax_year: Optional[int] = field(metadata=int_factory("/ReturnHeader/TaxYr"))
    us_address: Optional[str] = field(
        metadata=two_part_str_factory(
            "/IRS990/USAddress/AddressLine1Txt",
            "/IRS990/USAddress/AddressLine2Txt",
        )
    )
    us_city_name: Optional[str] = field(
        metadata=str_factory("/IRS990/USAddress/CityNm")
    )
    us_zip_code: Optional[str] = field(
        metadata=str_factory("/IRS990/USAddress/ZIPCd")
    )
    website_address: Optional[str] = field(
        metadata=str_factory("/IRS990/WebsiteAddressTxt")
    )

    def __init__(self, querier: Querier):
        for declaredField in fields(self):
            value = declaredField.metadata["factory"](querier)
            setattr(self, declaredField.name, value)

    def to_json(self) -> Dict[str, Any]:
        return asdict(self)


FilingFilter = Callable[[Filing], bool]



import re
from typing import Mapping

from .filing import Filing, FilingFilter
from .index import IndexFilter, IndexRecord


class FieldNotFound(Exception):
    name: str

    def __init__(self, name: str):
        self.name = name


def filter_filings(filters: Mapping[str, str],) -> FilingFilter:
    def _filter(filing: Filing) -> bool:
        passed = True
        for fieldName in filters:
            filter_text = filters[fieldName]
            if not hasattr(filing, fieldName):
                raise FieldNotFound(fieldName)
            value = getattr(filing, fieldName)
            if value is None:
                return False
            match = re.match(filter_text, str(value))
            if match is None:
                passed = False
                break
        return passed

    return _filter


def filter_index_record(filters: Mapping[str, str],) -> IndexFilter:
    def _filter(record: IndexRecord) -> bool:
        passed = True
        for field_name in filters:
            filter_text = filters[field_name]
            if not record.has_field(field_name):
                raise FieldNotFound(field_name)
            value = record.get_field(field_name)
            if value is None:
                return False
            match = re.match(filter_text, str(value))
            if match is None:
                passed = False
                break
        return passed

    return _filter



from logging import getLogger

from ._options import Options
from .annual_record import AnnualYear, get_annual_index
from .bmf_record import BMFRegion, get_bmf_index
from .constants import AWS_FILING_TEMPLATE, AWS_INDEX_TEMPLATE, IRS_BMF_TEMPLATE
from .downloader import HTTPDownloader
from .filters import filter_filings, filter_index_record
from .formatter import Formatter
from .index import Index
from .result import Result

# TODO: Split output methods into their own module (JSON, human-readable)

_logger = getLogger(__name__)


LIMIT_BEFORE_CONFIRM = 100
"""
If more than this number of documents will be downloaded, interactively
confirm before starting unless the --no-confirm flag was passed.
"""


def _message_count(download_count: int, cache_count: int) -> str:
    return (
        f"This would download {download_count} documents "
        + f"and process 0 total documents"
    )


def _confirm(
    download_count: int, cache_count: int, formatter: Formatter
) -> bool:
    # TODO: Give the formatter a crack at the prompt message
    message = _message_count(download_count, cache_count,)
    raw = input(message + ", continue? [yN] ",)
    return raw.strip().lower().startswith("y")


def run(options: Options):
    annual_downloader = HTTPDownloader(AWS_INDEX_TEMPLATE, options.index_cache,)
    bmf_downloader = HTTPDownloader(IRS_BMF_TEMPLATE, options.index_cache,)

    index = Index()
    for year in options.years:
        _logger.debug(f"applying year filter for {year}")
        typed_year = AnnualYear(int(year))
        index.add_annual_index(
            typed_year, get_annual_index(typed_year, annual_downloader,),
        )

    for region in options.regions:
        _logger.debug(f"applying region filter for {region}")
        typed_region = BMFRegion(region)
        index.add_bmf_index(
            typed_region, get_bmf_index(typed_region, bmf_downloader,),
        )

    if len(options.index_filters) > 0:
        _logger.debug(f"applying index filters: {options.index_filters}")
        index.add_filter(filter_index_record(options.index_filters))

    filing_downloader = HTTPDownloader(
        AWS_FILING_TEMPLATE, options.filing_cache,
    )
    result = Result(filing_downloader, index,)
    if len(options.filing_filters) > 0:
        _logger.debug(f"applying filing filters: {options.filing_filters}")
        result = result.add_filter(filter_filings(options.filing_filters))

    formatter = options.formatter
    download_count = result.download_count()

    # TODO: Have it check the cache and adjust downloads, process separately
    # TODO: Formatter should get a chance to modify / format messages

    if options.dry_run:
        _logger.debug(f"performing dry-run")
        print(_message_count(download_count, 0))
        return

    if not options.no_confirm and download_count > LIMIT_BEFORE_CONFIRM:
        _logger.debug(f"downloading {download_count} documents, confirming")
        if not _confirm(download_count, 0, formatter):
            _logger.debug(f"user aborted")
            # TODO: Use the formatter for this output
            print("Aborting")
            return

    formatter.prologue()
    formatter.write_all(result)
    formatter.epilogue()



from __future__ import annotations

import csv
from logging import getLogger
from typing import Any, Dict, Iterator, List, NamedTuple, NewType, TextIO

from .downloader import Downloader
from .types import EIN

_logger = getLogger(__name__)


ANNUAL_FIELD_NAMES: List[str] = [
    "RETURN_ID",
    "FILING_TYPE",
    "EIN",
    "TAX_PERIOD",
    "SUB_DATE",
    "TAXPAYER_NAME",
    "RETURN_TYPE",
    "DLN",
    "OBJECT_ID",
]


AnnualYear = NewType("AnnualYear", int)
"""
A filing year supported by the software. This will generally only
include 2017-2019, at least for now.
"""


AnnualRecordDict = NewType("AnnualRecordDict", Dict[str, str])
"""
A dictionary presumed to have the keys listed in `ANNUAL_FIELD_NAMES`
so that it can be used to create an instance of `AnnualRecord`.
"""


class AnnualRecord(NamedTuple):
    """
    A single record from an annual index (such as those hosted at
    https://registry.opendata.aws/irs990/) that can be used to get
    an object ID that will allow the actual filing to be retrieved
    from the internet.

    TODO: Audit the descriptions for correctness and completeness
    """

    @staticmethod
    def from_csv(index_file: TextIO,) -> Iterator[AnnualRecord]:
        for row in csv.DictReader(index_file):
            record = AnnualRecord.from_dict(AnnualRecordDict(row))
            yield record

    @staticmethod
    def from_dict(d: AnnualRecordDict,) -> AnnualRecord:
        params: Dict[str, str] = {}
        try:
            for name in ANNUAL_FIELD_NAMES:
                params[name.lower()] = d[name]
        except KeyError as e:
            _logger.error(f"malformed index row - '{e}' - '{d}'")
            raise
        return AnnualRecord(**params)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"AnnualRecord(taxpayer_name='{self.taxpayer_name}', ein='{self.ein}')"

    def to_dict(self) -> Dict[str, Any]:
        """
        Support for JSON serialization.
        """
        return dict(self.__dict__)

    return_id: str
    """Unique identifier for the tax filing."""

    filing_type: str
    """Type of filing, for example: 'EFILE'."""

    ein: str
    """Tax ID number of the organization."""

    tax_period: str
    """TODO: Find out what this is, exactly."""

    sub_date: str
    """Date the filing was submitted to the IRS."""

    taxpayer_name: str
    """Name of the organization that submitted the filing."""

    return_type: str
    """Type of form that was submitted, for example: '990'."""

    dln: str
    """TODO: Figure out what this is"""

    object_id: str
    """The unique identifier for this filing, used to download data."""


AnnualIndex = Dict[EIN, AnnualRecord]
"""
An Annual index as fetched from the IRS AWS bucket. It is representative
of all the filings received in the respective year, though it only
provides limited filtering ability because it doesn't include any
geographic information.
"""


# This is the global register for Annual indices. Since a given index
# should never change at runtime, we cache them for efficiency. Only
# the indices that are required will be created and they can be reused
# without significant additional cost.
_annual_indices: Dict[AnnualYear, AnnualIndex] = {}


def get_annual_index(year: AnnualYear, downloader: Downloader) -> AnnualIndex:
    """
    Get an AnnualIndex for the given year, using the given Downloader
    to retrieve the appropriate file. The provided Downloader must be
    able to accept a four-digit year as its document key.

    >>> from pyrs990.downloader import FakeDownloader
    >>> documents = {"2018": open("fixtures/annual_index.csv")}
    >>> downloader = FakeDownloader(documents)
    >>> index0 = get_annual_index(AnnualYear(2018), downloader)
    >>> index1 = get_annual_index(AnnualYear(2018), downloader)
    >>> id(index0) == id(index1)
    True
    >>> len(index0)
    2
    >>> "133085892" in index0
    True
    """
    if year in _annual_indices:
        return _annual_indices[year]

    index: AnnualIndex = {}
    _annual_indices[year] = index

    index_file = downloader.fetch(str(year))
    for record in AnnualRecord.from_csv(index_file):
        index[EIN(record.ein)] = record
    return index



import logging
from io import StringIO
from typing import Dict, Iterable, TextIO

import requests

from .cache import Cache

_logger = logging.getLogger(__name__)


class DownloaderException(Exception):
    pass


class Downloader:
    """
    A class responsible for fetching documents based on an identifier.
    """

    def fetch(self, document: str,) -> TextIO:
        raise NotImplementedError()

    def fetch_all(self, documents: Iterable[str],) -> Iterable[TextIO]:
        raise NotImplementedError()


class FakeDownloader(Downloader):
    """
    A fake for testing pieces of the package that rely on a Downloader
    implementation. Give it a dictionary of document keys (strings)
    and document contents (file-like objects) and it does the rest.
    """

    _documents: Dict[str, TextIO]

    def __init__(self, documents: Dict[str, TextIO]):
        self._documents = documents

    def fetch(self, document: str,) -> TextIO:
        return self._documents[document]

    def fetch_all(self, documents: Iterable[str],) -> Iterable[TextIO]:
        return [self.fetch(d) for d in documents]


class HTTPDownloader(Downloader):
    """
    A downloader implementation that uses a template URL to fetch documents
    using HTTP. The cache is checked before any requests are made.

    The cache will always be checked before a network request is made.

    The `url_template` passed to the constructor should be a valid Python
    format string. The only context made available to it will be a key called
    `document` which will match the object ID of the filing to be
    downloaded, or the year of the index to be downloaded.

    If the download fails for any reason a `DownloaderException` will be
    raised. Its message will include the body of the response.

    TODO: Remove "requests" dependency and just use the std lib
    """

    _cache: Cache

    _url_template: str

    def __init__(
        self, url_template: str, cache: Cache,
    ):
        self._cache = cache
        self._url_template = url_template

    def fetch(self, document: str,) -> TextIO:
        _logger.info(f"fetching document '{document}'")
        content = self._cache.get(document)
        if content is None:
            url = self._url_template.format(document=document)
            _logger.info(f"downloading '{url}'")
            response = requests.get(url)
            _logger.info(f"download finished '{response.reason}'")

            # Encoding sniffing works way better than just trusting
            # the Content-Type header
            response.encoding = response.apparent_encoding

            # We're pretty loose with what we assume to be a successful
            # download here but that should be fine for now.
            # TODO: Decide how to handle this exception
            if response.status_code < 200 or response.status_code > 299:
                _logger.warning(f"download failed '{response.status_code}'")
                raise DownloaderException(response.text)

            content = self._cache.put(document, response.text)
        else:
            _logger.info("found document in cache")

        return StringIO(content)

    def fetch_all(self, documents: Iterable[str],) -> Iterable[TextIO]:
        for document in documents:
            yield self.fetch(document)



AWS_FILING_TEMPLATE = (
    "https://s3.amazonaws.com/irs-form-990/{document}_public.xml"
)
"""
A URL template that uses the AWS S3 bucket maintained by the IRS. For
use with `HTTPDownloader` to download filing documents.
"""

AWS_INDEX_TEMPLATE = (
    "https://s3.amazonaws.com/irs-form-990/index_{document}.csv"
)
"""
A URL template that uses the AWS S3 bucket maintained by the IRS. For
use with the `HTTPDownloader` to download index documents.
"""

CURRENT_VERSION = "0.0.1"

DEFAULT_CACHE_PATH = ".pyrs990-cache"

IRS_BMF_TEMPLATE = "https://www.irs.gov/pub/irs-soi/eo_{document}.csv"
"""
A URL template for downloading the Business Master File data from
the IRS web site. This provides us with location data without
having to download every single filing.
"""

PROGRAM_DESCRIPTION = (
    "A utility to extract and serialize IRS Form 990 "
    + "data on non-profit organizations."
)

PROGRAM_NAME = "pyrs990"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrs990']

package_data = \
{'': ['*']}

install_requires = \
['defusedxml>=0.6.0,<0.7.0', 'requests>=2.23.0,<3.0.0']

entry_points = \
{'console_scripts': ['pyrs990 = pyrs990']}

setup_kwargs = {
    'name': 'pyrs990',
    'version': '0.1.2',
    'description': 'A tool for fetching and filtering IRS 990 data.',
    'long_description': '![PyRS990 Header](https://github.com/code-for-montana/pyrs990/raw/master/pyrs990_header.png)\n\nIt\'s a pun. Get it?\n\nA Python application and library that can grab all sorts of IRS Form 990\ndata on non-profit organizations and put it into a format that can be\nconsumed easily by other applications.\n\n## Up and Running\n\nThe instructions below should allow you to get the software working\nfor your purpose (user or developer). If you run into trouble please,\nplease let us know so that we can update the instructions (or fix the\nbug you ran into).\n\n### User\n\nFor now you need to clone the repo to use it. Eventually we\'ll package it.\n\n  1. Make sure you have Python 3.8 available\n  1. Install [Poetry](https://python-poetry.org/) if you don\'t already have it\n  1. Clone the whole repo, `cd` into the `pyrs990` directory\n  1. Install dependencies - `poetry install`\n  1. Run it, some very simple examples are below:\n      1. `poetry run python -m pyrs990 --zip 59801 --use-disk-cache`\n      1. ...more examples coming soon\n  1. Run the commands again, notice the cache speedup\n  1. The cache is set to `./.pyrs990-cache/`\n\n### Developer\n\nThis project uses [Poetry](https://python-poetry.org/) because it\'s pretty slick\nand does a lot of stuff automatically and the developers are not usually Python\npeople, so that\'s great!\n\n  1. Make sure you have Python 3.8 available\n  1. Install [Poetry](https://python-poetry.org/) if you don\'t already have it\n  1. Clone the whole repo, `cd` into the `pyrs990` directory\n  1. Install dependencies - `poetry install`\n  1. If you need to add dependencies:\n      1. `poetry add coolpkg`\n  1. Make a pull request!\n\n## About the Data\n\nRight now we pull data that originated with the IRS (hence the silly name)\nbut we get it from a couple sources and information about what is actually\navailable is a little spread out as well.\n\n### Structure\n\nThere are two indices used to narrow down the list of filing documents\nthat must be downloaded a satisfy a given query. The first is an\nannual index (we refer to it as "Annual" or "Annual Index" in the\ncode). This index contains all filings processed by the IRS for a\ngiven calendar year.\n\nNote that this does not necessarily have anything\nto do with the filing year. An organization might, for example, file\nits 2016 990 in either 2017 or 2018 (or even later). There is a field,\ndescribed below, called `tax_period` that reflects the filing period.\nIn the future, we intend to further abstract this so that it is\neasier to use.\n\nThe annual index also contains a field called `object_id` that tells\nus where to find the XML document that corresponds to that row in\nthe index. PyRS990 abstracts this away, but it is still good to be\naware of it.\n\nThe second index is the "Exempt Organizations Business Master File"\ndistributed by the IRS. We refer to it as the "BMF Index". This\nindex provides the physical address of each organization, along\nwith some other helpful information. This allows the data to be\nqueried by state, zip code, and so on, which greatly reduces the\nnumber of filing documents that must be downloaded for many queries.\n\nIndices may be used to query filing documents from the command\nline using various options. Note that there are options for both\nindices and for the filing documents themselves. If possible, it\nis a good idea to try to use as many index fields as you can to\nreduce the number of files you have to download.\n\nSee the example queries for more information.\n\n### Sources\n\nThe [IRS BMF index files](https://www.irs.gov/charities-non-profits/exempt-organizations-business-master-file-extract-eo-bmf)\nare hosted by the IRS directly and are available by state and region.\n\n[Descriptions of the variables](https://www.irs.gov/pub/irs-soi/eo_info.pdf)\ncontained in the files and the process used to build them are\nalso available (it is also linked from the page above).\n\nThe annual index files come from an\n[AWS S3 bucket](https://registry.opendata.aws/irs990/)\nmanaged by the IRS. The contents of the bucket are described there.\n\nThere is also [a readme](https://docs.opendata.aws/irs-990/readme.html)\nthat demonstrates how to download the files here (it is also linked\nfrom the page above):\n\nThe filing documents themselves also come from this same\n[AWS S3 bucket](https://registry.opendata.aws/irs990/)\nin XML format. For the extremely XML-savvy, you can checked out the\n[schema documentation](https://www.irs.gov/e-file-providers/current-valid-xml-schemas-and-business-rules-for-exempt-organizations-modernized-e-file)\non the IRS website. PyRS990 abstracts this away, however,\nso there\'s no real need to understand it if you only want to access the\ndata in a convenient format.\n\nFinally, while not strictly a data source, the\n[IRSx documentation](http://www.irsx.info/) created\nby ProPublica contains descriptions of many of the filing fields in a\nsimple, readable format. For developers, PyRS990 has been designed to\nwork with the exact XPath selectors listed in the IRSx documentation, so\nif you want to add a field to the `Filing` object, this is the place to\nlook first.\n',
    'author': 'George Lesica',
    'author_email': 'george@lesica.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/code-for-montana/pyrs990',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import Iterator, NamedTuple

from ._options import Options


class Config(NamedTuple):
    """
    A container for global configuration options to allow hierarchical
    specification through a configuration file, environment variables,
    and built-in constant values.
    """

    @staticmethod
    def config_file_paths() -> Iterator[str]:
        # check system global path
        # check home directory
        # check CWD
        # check directories above CWD???
        pass

    @staticmethod
    def from_options(options: Options) -> Config:
        pass
